
import os
import unicodedata

from .constants import CHAR_LOWERCASE, CHAR_NUMERIC, CHAR_UPPERCASE, \
    VALID_PATTERN_MIN, RECIPE_NAME, VALID_RECIPE_MIN, PATTERN_NAME, \
    WORKFLOW_NAME, STEP_NAME, VARIABLES_NAME, MEOW_MODE, CWL_MODE, \
    VALID_WORKFLOW_MIN, VALID_STEP_MIN, VALID_SETTING_MIN, \
    VALID_PATTERN_OPTIONAL, VALID_RECIPE_OPTIONAL, VALID_SETTING_OPTIONAL, \
    VALID_STEP_OPTIONAL, VALID_WORKFLOW_OPTIONAL, CWL_CLASS_WORKFLOW, \
    CWL_CLASS_COMMAND_LINE_TOOL, CWL_CLASS, TRIGGER_PATHS, TRIGGER_RECIPES, \
    INPUT_FILE, VALID_SWEEP_MIN, VALID_SWEEP_OPTIONAL, SWEEP_START, \
    SWEEP_STOP, SWEEP_JUMP, CHAR_LINES, ENVIRONMENTS, ENVIRONMENTS_MIG, \
    ENVIRONMENTS_LOCAL, VALID_ENVIRONMENT_TYPES, VALID_ENVIRONMENTS_MIG, \
    VALID_ENVIRONMENTS_LOCAL, ENVIRONMENTS_LOCAL_VERSION, \
    ENVIRONMENTS_LOCAL_DEPENDENCIES, CHAR_COMPARISON, \
    ENVIRONMENTS_MIG_RUNTIME_ENVIRONMENTS, ENVIRONMENTS_MIG_RETRIES, \
    ENVIRONMENTS_MIG_NOTIFICATION, ENVIRONMENTS_MIG_ENVIRONMENT_VARIABLES, \
    ENVIRONMENTS_MIG_DISKS, ENVIRONMENTS_MIG_FILL, ENVIRONMENTS_MIG_NODES, \
    ENVIRONMENTS_MIG_MEMORY, ENVIRONMENTS_MIG_CPU_ARCHITECTURE, \
    ENVIRONMENTS_MIG_WALL_TIME, ENVIRONMENTS_MIG_CPU_CORES, \
    VALID_ENVIRONMENTS_MIG_FILLS, COMPARITORS


def is_a_number(string):
    """
    Helper function to determine if a given string expresses a number

    :param string: (str) The string to check

    :return: (bool) Will return True if given string expresses a number, and
    False otherwise
    """
    check_input(string, 'str', 'string')

    try:
        float(string)
        return True
    except ValueError:
        pass

    try:
        unicodedata.numeric(string)
        return True
    except (TypeError, ValueError):
        pass

    return False


def check_input(variable, expected_type, name, or_none=False):
    """
    Checks if a given variable is of the expected type. Raises TypeError or
    ValueError as appropriate if any issues are encountered.

    :param variable: (any) variable to check type of

    :param expected_type: (type) expected type of the provided variable

    :param name: (str) name of the variable, used to make clearer debug
    messages.

    :param or_none: (optional) boolean of if the variable can be unset.
    Default value is False.

    :return: No return.
    """

    if not variable and expected_type is not bool and or_none is False:
        raise ValueError('variable %s was not given' % name)

    if not expected_type:
        raise ValueError('\'expected_type\' %s was not given' % expected_type)

    if not or_none:
        if not isinstance(variable, expected_type):
            raise TypeError(
                'Expected %s type was %s, got %s'
                % (name, expected_type, prep_html(type(variable)))
            )
    else:
        if not isinstance(variable, expected_type) \
                and not isinstance(variable, type(None)):
            raise TypeError(
                'Expected %s type was %s or None, got %s'
                % (name, expected_type, prep_html(type(variable)))
            )


def is_valid_list_entry(arg, valid_entries, name):
    """
    Checks that a given args ia in a given list. Will raise an exception if
    arg is not in valid_entries.

    :param arg: (object) The object to check.

    :param valid_entries: (list) The list to check against.

    :param name: (str) The name of the variable. Only used for error printing.

    :return: No return.
    """
    check_input(valid_entries, list, name)
    if arg not in valid_entries:
        raise ValueError('Invalid entry for %s. %s is not in %s'
                         % (name, arg, valid_entries))


def check_input_args(args, valid_args):
    """
    Checks that given args are allowed. Raises ValueError or TypeError as
    appropriate if provided arguments are not valid.

    :param args: (dict) arguments to check.

    :param valid_args: (dict) arguments that are expected and their type.

    :return: No return
    """
    if not isinstance(args, dict):
        raise TypeError("Arguments provided in invalid format")

    for key, arg in args.items():
        if key not in valid_args:
            raise ValueError(
                "Unsupported argument %s. Valid are: %s. "
                % (key, list(valid_args.keys()))
            )
        if not isinstance(arg, valid_args[key]):
            raise TypeError(
                'Argument %s is in an unexpected format. Expected %s but got '
                '%s' % (key, valid_args[key], prep_html(type(arg)))
            )


def valid_string(variable, name, valid_chars):
    """
    Checks that all characters in a given string are present in a provided
    list of characters. Will raise an ValueError if unexpected character is
    encountered.

    :param variable: (str) variable to check.

    :param name: (str) name of variable to check. Only used to clarify debug
    messages.

    :param valid_chars: (str) collection of valid characters.

    :return: No return.
    """
    check_input(variable, str, name)
    check_input(valid_chars, str, 'valid_chars')

    for char in variable:
        if char not in valid_chars:
            raise ValueError(
                "Invalid character %s in %s '%s'. Only valid characters are: "
                "%s" % (char, name, variable, valid_chars)
            )


def valid_dir_path(path, name):
    """
    Checks that a given string is a valid path. Raises ValueError if not a
    valid path.

    :param path: (str) path to check.

    :param name: (str) name of variable to check. Only used to clarify debug
    messages.

    :return: No return.
    """
    check_input(path, str, name)

    valid_chars = \
        CHAR_NUMERIC + CHAR_UPPERCASE + CHAR_LOWERCASE + '-_.' + os.path.sep

    for char in path:
        if char not in valid_chars:
            raise ValueError(
                'Invalid character %s in string %s for variable %s. Only '
                'valid characters are %s' % (char, path, name, valid_chars)
            )


def dir_exists(path, create=False):
    """
    Checks a given path exists and is a directory. Raises a ValueError if the
    path either does not exist or is not a directory.

    :param path: (str) the path to check

    :param create: (bool) toggle to create dir if it does not exist. Default
    is false

    :return: (no return)
    """
    if not os.path.exists(path):
        if create:
            os.mkdir(path)
        else:
            raise ValueError("Directory '%s' does not exist. " % path)
    if not os.path.isdir(path):
        raise ValueError("Path '%s' is not a directory. " % path)


def valid_file_path(path, name, extensions=None):
    """
    Checks that a given string is a valid file path. Raises ValueError if not
    a valid path.

    :param path: (str) path to check.

    :param name: (str) name of variable to check. Only used to clarify debug
    messages.

    :param extensions: (list)[optional]. List of possible extensions to
    check in path. Defaults to None. Extensions should be of the form '.txt'.

    :return: No return.
    """
    check_input(path, str, name)
    check_input(extensions, list, 'extensions', or_none=True)

    valid_chars = \
        CHAR_NUMERIC + CHAR_UPPERCASE + CHAR_LOWERCASE + '-_.' + os.path.sep

    if path.count('.') > 1:
        raise ValueError(
            "Too many '.' characters in %s path. Should only be one. " % name
        )
    if path.count('.') == 0:
        raise ValueError(
            "No file extension found in %s path. Please define one. " % name
        )

    if extensions:
        extension = path[path.index('.'):]
        if extension not in extensions:
            raise ValueError(
                '%s is not a supported format for variable %s. Please only '
                'use one of the following: %s. '
                % (extension, name, extensions)
            )

    for char in path:
        if char not in valid_chars:
            raise ValueError(
                'Invalid character %s in string %s for variable %s. Only '
                'valid characters are %s' % (char, path, name, valid_chars)
            )


def valid_param_sweep(to_test, name):
    """
    Checks that a given dict is a valid Pattern parameter sweep definition.
    Raises ValueError if a problem is encountered.

    :param to_test: (dict) The dictionary to test.

    :param name: (str) The name of the given dictionary. Used for ease of
    debugging.

    :return: No return.
    """
    check_input(to_test, dict, 'parameter sweep')
    check_input(name, str, 'parameter sweep name')
    valid, msg = is_valid_dict(
        to_test,
        VALID_SWEEP_MIN,
        VALID_SWEEP_OPTIONAL,
        'parameter sweep',
        MEOW_MODE,
        strict=True
    )

    if not valid:
        raise ValueError(msg)

    if not isinstance(to_test[SWEEP_START], type(to_test[SWEEP_STOP])) or \
            not isinstance(to_test[SWEEP_START], type(to_test[SWEEP_JUMP])):
        raise ValueError(
            'All parameter sweep values should be of the same type. Found '
            'types were start: %s, stop: %s, and jump: %s'
            % (prep_html(type(to_test[SWEEP_START])),
               prep_html(type(to_test[SWEEP_STOP])),
               prep_html(type(to_test[SWEEP_JUMP])))
        )

    # Try to check that this loop is not infinite
    if to_test[SWEEP_JUMP] == 0:
        raise ValueError(
            'Cannot create parameter sweep with a jump value of zero, as '
            'this would create an infinite loop'
        )
    elif to_test[SWEEP_JUMP] > 0:
        if not to_test[SWEEP_STOP] > to_test[SWEEP_START]:
            raise ValueError(
                'Cannot create parameter sweep with a positive jump value '
                'where the end point is smaller than the start. '
            )
    elif to_test[SWEEP_JUMP] < 0:
        if not to_test[SWEEP_STOP] < to_test[SWEEP_START]:
            raise ValueError(
                'Cannot create parameter sweep with a negative jump value '
                'where the end point is larger than the start. '
            )


def valid_recipe_name(name):
    """
    Validates that a given name is a valid recipe name.

    :param name: (str) The name to test.

    :return: No return
    """
    valid_string(
        name,
        'recipe name',
        CHAR_UPPERCASE
        + CHAR_LOWERCASE
        + CHAR_NUMERIC
        + CHAR_LINES
    )


def valid_pattern_name(name):
    """
    Validates that a given name is a valid pattern name.

    :param name: (str) The name to test.

    :return: No return
    """
    valid_string(
        name,
        'pattern name',
        CHAR_UPPERCASE
        + CHAR_LOWERCASE
        + CHAR_NUMERIC
        + CHAR_LINES
    )


def is_valid_dict(to_test, required_args, optional_args, name, paradigm,
                  strict=False):
    """
    Validates that a given dict has the expected arguments.

    :param to_test: (dict) the dictionary whose arguments are to be checked.

    :param required_args: (dict) A dictionary of expected arguments and the
    types of those arguments. Check will fail if all of these are not provided.

    :param optional_args: (dict) A dictionary of possible arguments and the
    types of those arguments. These will be type checked if provided, but are
    not necessary for a 'to_test' dict to be valid.

    :param name: (str) The name of this dict type. Used for debugging messages.

    :param paradigm: (str) The paradigm the tested dict is operating within.
    Only used for debugging messages.

    :param strict: (bool)[optional] Option to be strict about arguments. If
    True then any extra arguments that have been provided will fail. Default
    is False.

    :return: (Tuple(bool, str)) First value is boolean. True = to_test
    is valid, False = to_test is not valid. Second value is feedback
    string and will be empty if first value is True.
    """

    if not to_test:
        return False, 'A %s %s was not provided. ' % (paradigm, name)

    if not isinstance(to_test, dict):
        return False, \
               'The %s %s was incorrectly formatted. Should be a dict, but ' \
               '%s is a %s' \
               % (paradigm, name, to_test, prep_html(type(to_test)))

    message = 'The %s %s %s had an incorrect structure, ' \
              % (paradigm, name, to_test)
    for key, value in required_args.items():
        if key not in to_test:
            message += 'it is missing key %s. ' % key
            return False, message
        if isinstance(value, list):
            if type(to_test[key]) not in value:
                message += \
                    ' %s is expected to have types %s but actually has %s. ' \
                    % (key, prep_html(value), prep_html(type(to_test[key])))
                return False, message
        else:
            if not isinstance(to_test[key], value):
                message += \
                    ' %s is expected to have type %s but actually has %s. ' \
                    % (key, prep_html(value), prep_html(type(to_test[key])))
                return False, message

    for key, value in optional_args.items():
        if key in to_test:
            if isinstance(value, list):
                if type(to_test[key] not in value):
                    message += \
                        ' %s is expected to have types %s but actually has ' \
                        '%s. ' \
                        % (to_test[key], prep_html(value),
                           prep_html(type(to_test[key])))
                    return False, message
            else:
                if not isinstance(to_test[key], value):
                    message += \
                        ' %s is expected to have type %s but actually has ' \
                        '%s. ' \
                        % (to_test[key], prep_html(value),
                           prep_html(type(to_test[key])))
                    return False, message

    if strict:
        for key in to_test.keys():
            if key not in required_args and key not in optional_args:
                message += ' contains extra key %s' % key
                return False, message
    return True, ''


def is_valid_pattern_dict(to_test, strict=False):
    """
    Validates that the passed dictionary can be used to create a new Pattern
    object.

    :param to_test: (dict) object to be tested.

    :param strict: (bool)[optional] Option to be strict about arguments. If
    True then any extra arguments that have been provided will fail. Default
    is False.

    :return: (Tuple(bool, str)) First value is boolean. True = to_test
    is Pattern, False = to_test is not Pattern. Second value is feedback
    string and will be empty if first value is True.
    """

    valid, msg = is_valid_dict(
        to_test,
        VALID_PATTERN_MIN,
        VALID_PATTERN_OPTIONAL,
        PATTERN_NAME,
        MEOW_MODE,
        strict=strict
    )

    if not valid:
        return False, msg

    if not to_test[TRIGGER_PATHS]:
        return False, "%s invalid: %s. Must be defined" \
               % (TRIGGER_PATHS, to_test[TRIGGER_PATHS])

    if not to_test[INPUT_FILE]:
        return False, "%s invalid: %s. Must be defined" \
               % (INPUT_FILE, to_test[INPUT_FILE])

    if TRIGGER_RECIPES not in to_test:
        return False, "'trigger_recipes' key was not in %s. " \
               % str(list(to_test.keys()))
    if not isinstance(to_test[TRIGGER_RECIPES], dict):
        return False, \
               "Trigger id's have not be stored in the correct format. " \
               "Expected dict but got %s." \
               % prep_html(type(to_test['trigger_recipes']))

    for trigger_id, trigger in to_test[TRIGGER_RECIPES].items():
        if not isinstance(trigger_id, str):
            return False, "Trigger id %s is a %s, not the expected str." \
                   % (str(trigger_id), prep_html(type(trigger_id)))
        if not isinstance(trigger, dict):
            return False, "Trigger %s is a %s, not the expected dict." \
                   % (str(trigger), prep_html(type(trigger)))

        if not trigger:
            return False, "Trigger is empty. Should contain at least a " \
                          "single recipe"

        for rec_id, recipe in trigger.items():
            if not isinstance(rec_id, str):
                return False, "Recipe id %s is a %s, not the expected str." \
                       % (str(rec_id), prep_html(type(rec_id)))
            valid, msg = is_valid_recipe_dict(recipe)
            if not valid:
                return False, msg

    return True, ''


def is_valid_recipe_dict(to_test, strict=False):
    """
    Validates that the passed dictionary expresses a meow recipe.

    :param to_test: (dict) A dictionary, hopefully expressing a meow recipe

    :param strict: (bool)[optional] Option to be strict about arguments. If
    True then any extra arguments that have been provided will fail. Default
    is False.

    :return: (Tuple(bool, str)) First value is boolean. True = to_test
    is recipe, False = to_test is not recipe. Second value is feedback
    string and will be empty if first value is True.
    """

    # Note that recipe may be an empty dict if a recipe has not yet been
    # registered on the mig
    if isinstance(to_test, dict) and not to_test:
        return True, ''

    status, feedback = is_valid_dict(
        to_test,
        VALID_RECIPE_MIN,
        VALID_RECIPE_OPTIONAL,
        RECIPE_NAME,
        MEOW_MODE,
        strict=strict
    )

    if not status:
        return status, feedback

    if ENVIRONMENTS in to_test:
        return is_valid_environments_dict(to_test[ENVIRONMENTS], strict=strict)
    else:
        return status, feedback


def is_valid_environments_dict(to_test, strict=False):
    # Note that recipe environments definition may be an empty dict
    if isinstance(to_test, dict) and not to_test:
        return True, ''

    for env, defs in to_test.items():
        check_input(env, str, 'environment key')
        check_input(defs, dict, 'environment values')

        if env == ENVIRONMENTS_MIG:
            return is_valid_mig_environment(to_test[env], strict=strict)
        elif env == ENVIRONMENTS_LOCAL:
            return is_valid_local_environment(to_test[env], strict=strict)
        else:
            if strict:
                return False, "Unknown environment type '%s'. Valid are: %s" \
                       % (env, VALID_ENVIRONMENT_TYPES)


def is_valid_mig_environment(to_test, strict=False):
    if isinstance(to_test, dict) and not to_test:
        return True, ''

    status, feedback = is_valid_dict(
        to_test,
        {},
        VALID_ENVIRONMENTS_MIG,
        ENVIRONMENTS_MIG,
        MEOW_MODE)

    if not status:
        return status, feedback

    if strict:
        for k, v in to_test.items():
            if k not in VALID_ENVIRONMENTS_MIG:
                return False, "Unknown dependency '%s'. Valid are: %s" \
                       % (k, VALID_ENVIRONMENTS_MIG)

    if ENVIRONMENTS_MIG_NODES in to_test:
        valid_string(
            to_test[ENVIRONMENTS_MIG_NODES],
            ENVIRONMENTS_MIG_NODES,
            CHAR_NUMERIC
        )
    if ENVIRONMENTS_MIG_CPU_CORES in to_test:
        valid_string(
            to_test[ENVIRONMENTS_MIG_CPU_CORES],
            ENVIRONMENTS_MIG_CPU_CORES,
            CHAR_NUMERIC
        )
    if ENVIRONMENTS_MIG_WALL_TIME in to_test:
        valid_string(
            to_test[ENVIRONMENTS_MIG_WALL_TIME],
            ENVIRONMENTS_MIG_WALL_TIME,
            CHAR_NUMERIC
        )
    if ENVIRONMENTS_MIG_MEMORY in to_test:
        valid_string(
            to_test[ENVIRONMENTS_MIG_MEMORY],
            ENVIRONMENTS_MIG_MEMORY,
            CHAR_NUMERIC
        )
    if ENVIRONMENTS_MIG_DISKS in to_test:
        valid_string(
            to_test[ENVIRONMENTS_MIG_DISKS],
            ENVIRONMENTS_MIG_DISKS,
            CHAR_NUMERIC
        )
    if ENVIRONMENTS_MIG_CPU_ARCHITECTURE in to_test:
        valid_string(
            to_test[ENVIRONMENTS_MIG_CPU_ARCHITECTURE],
            ENVIRONMENTS_MIG_CPU_ARCHITECTURE,
            CHAR_NUMERIC + CHAR_UPPERCASE + CHAR_LOWERCASE + CHAR_LINES
        )
    if ENVIRONMENTS_MIG_FILL in to_test:
        for fill in to_test[ENVIRONMENTS_MIG_FILL]:
            if fill not in VALID_ENVIRONMENTS_MIG_FILLS:
                return False, "Invalid fill '%s'. Valid are: %s" \
                       % (fill, VALID_ENVIRONMENTS_MIG_FILLS)

    if ENVIRONMENTS_MIG_ENVIRONMENT_VARIABLES in to_test:
        for env_var in to_test[ENVIRONMENTS_MIG_ENVIRONMENT_VARIABLES]:
            valid_string(
                env_var,
                'environment variable',
                CHAR_NUMERIC + CHAR_UPPERCASE + CHAR_LOWERCASE + CHAR_LINES
                + '= '
            )
            if ' ' in env_var:
                env_var = env_var.replace(' ', '')
            if env_var.count('=') != 1:
                return False, "Invalid assignment in '%s'" % env_var
            values = env_var.split('=')
            if len(values) != 2 or not values[0] or not values[1]:
                return False, "Invalid assignment form for '%s'" % env_var

    if ENVIRONMENTS_MIG_NOTIFICATION in to_test:
        for notification in to_test[ENVIRONMENTS_MIG_NOTIFICATION]:
            valid_string(
                notification,
                'notification',
                CHAR_NUMERIC + CHAR_UPPERCASE + CHAR_LOWERCASE + CHAR_LINES
                + ': @.\/+'
            )
            if ' ' in notification:
                notification = notification.replace(' ', '')
            if notification.count(':') != 1:
                return False, "Invalid assignment in '%s'" % notification
            values = notification.split(':')
            if len(values) != 2 or not values[0] or not values[1]:
                return False, "Invalid assignment form for '%s'" % notification

    if ENVIRONMENTS_MIG_RETRIES in to_test:
        valid_string(
            to_test[ENVIRONMENTS_MIG_RETRIES],
            ENVIRONMENTS_MIG_RETRIES,
            CHAR_NUMERIC
        )
    if ENVIRONMENTS_MIG_RUNTIME_ENVIRONMENTS in to_test:
        for runtime_env in to_test[ENVIRONMENTS_MIG_RUNTIME_ENVIRONMENTS]:
            valid_string(
                runtime_env,
                'runtime environment',
                CHAR_NUMERIC + CHAR_UPPERCASE + CHAR_LOWERCASE + CHAR_LINES
            )

    return True, ''


def is_valid_local_environment(to_test, strict=False):
    if isinstance(to_test, dict) and not to_test:
        return True, ''

    status, feedback = is_valid_dict(
        to_test,
        {},
        VALID_ENVIRONMENTS_LOCAL,
        ENVIRONMENTS_LOCAL,
        MEOW_MODE)

    if not status:
        return status, feedback

    if strict:
        for k, v in to_test.items():
            if k not in VALID_ENVIRONMENTS_LOCAL:
                return False, "Unknown dependency '%s'. Valid are: %s" \
                       % (k, VALID_ENVIRONMENTS_LOCAL)

    if ENVIRONMENTS_LOCAL_DEPENDENCIES in to_test:
        for dependency in to_test[ENVIRONMENTS_LOCAL_DEPENDENCIES]:
            valid_string(
                dependency,
                'dependency',
                CHAR_UPPERCASE + CHAR_LOWERCASE + CHAR_NUMERIC + CHAR_LINES
                + '.>='
            )

            matches = [x for x in COMPARITORS if x in dependency]
            if not matches:
                pass
            elif len(matches) == 1:
                match = matches[0]
                values = dependency.split(match)
                if len(values) != 2 or not values[0] or not values[1]:
                    return False, "Invalid assignment form for '%s'" \
                           % dependency
                valid_string(
                    values[0],
                    'dependency package',
                    CHAR_UPPERCASE + CHAR_LOWERCASE + CHAR_NUMERIC + CHAR_LINES
                )
                valid_string(
                    values[1],
                    'dependency package version',
                    CHAR_NUMERIC + '.'
                )
            else:
                return False, "Invalid assignment in '%s'" % dependency

    return True, ''


def is_valid_workflow_dict(to_test, strict=False):
    """
    Validates that the passed dictionary expresses a cwl workflow.

    :param to_test: (dict) A dictionary, hopefully expressing a cwl workflow

    :param strict: (bool)[optional] Option to be strict about arguments. If
    True then any extra arguments that have been provided will fail. Default
    is False.

    :return: (Tuple(bool, str)) First value is boolean. True = to_test
    is workflow, False = to_test is not workflow. Second value is feedback
    string and will be empty if first value is True.
    """
    valid, msg = is_valid_dict(
        to_test,
        VALID_WORKFLOW_MIN,
        VALID_WORKFLOW_OPTIONAL,
        WORKFLOW_NAME,
        CWL_MODE,
        strict=strict
    )

    if not valid:
        return False, msg

    if to_test[CWL_CLASS] != CWL_CLASS_WORKFLOW:
        msg = "%s class is '%s' not %s" \
              % (WORKFLOW_NAME, to_test[CWL_CLASS], CWL_CLASS_WORKFLOW)
        return False, msg

    return True, ''


def is_valid_step_dict(to_test, strict=False):
    """
    Validates that the passed dictionary expresses a cwl step.

    :param to_test: (dict) A dictionary, hopefully expressing a cwl step

    :param strict: (bool)[optional] Option to be strict about arguments. If
    True then any extra arguments that have been provided will fail. Default
    is False.

    :return: (Tuple(bool, str)) First value is boolean. True = to_test
    is step, False = to_test is not step. Second value is feedback
    string and will be empty if first value is True.
    """
    valid, msg = is_valid_dict(
        to_test,
        VALID_STEP_MIN,
        VALID_STEP_OPTIONAL,
        STEP_NAME,
        CWL_MODE,
        strict=strict
    )

    if not valid:
        return False, msg

    if to_test[CWL_CLASS] != CWL_CLASS_COMMAND_LINE_TOOL:
        return False, "%s class is '%s' not %s" \
               % (STEP_NAME, to_test[CWL_CLASS], CWL_CLASS_COMMAND_LINE_TOOL)

    return True, ''


def is_valid_setting_dict(to_test, strict=False):
    """
    Validates that the passed dictionary expresses cwl arguments.

    :param to_test: (dict) A dictionary, hopefully expressing cwl arguments

    :param strict: (bool)[optional] Option to be strict about arguments. If
    True then any extra arguments that have been provided will fail. Default
    is False.

    :return: (function call to 'is_valid_dict'). Returns a function call to
    'is_valid_dict'.
    """
    return is_valid_dict(
        to_test,
        VALID_SETTING_MIN,
        VALID_SETTING_OPTIONAL,
        VARIABLES_NAME,
        CWL_MODE,
        strict=strict
    )


def prep_html(prep):
    """
    Helper function to strip the pointy brackets out of type strings so that
    their contents appears in html

    :param prep: (type) or (list) An object type, or a list of object types

    :return: (str) The finalised string
    """
    if isinstance(prep, list):
        for elem in prep:
            if not isinstance(elem, type):
                raise TypeError(
                    "'prep_html given incorrect format %s when expected 'type'"
                    % prep_html(type(elem))
                )
    else:
        if not isinstance(prep, type):
            raise TypeError(
                "'prep_html given incorrect format %s when expected 'type'"
                % prep_html(type(prep))
            )

    string = str(prep)
    string = string.replace('>', '')
    string = string.replace('<', '')

    return string




# This code is heavily based off the 'mig/server/grid_events.py' file
# contained in the MiG source code at: https://sourceforge.net/projects/migrid/

import copy
import glob
import os
import time
import shutil
import re
import fnmatch
import subprocess
import threading
import pkg_resources

from datetime import datetime
from multiprocessing import Process, Pipe, current_process
from multiprocessing.connection import wait
from random import SystemRandom
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler, FileCreatedEvent, \
    FileModifiedEvent, FileDeletedEvent, DirCreatedEvent, DirModifiedEvent, \
    DirDeletedEvent

from .constants import PATTERNS, RECIPES, NAME, SOURCE, CHAR_LOWERCASE, \
    CHAR_UPPERCASE, CHAR_NUMERIC, RECIPE, KEYWORD_DIR, KEYWORD_EXTENSION, \
    KEYWORD_FILENAME, KEYWORD_JOB, KEYWORD_PATH, KEYWORD_PREFIX, \
    KEYWORD_REL_DIR, KEYWORD_REL_PATH, KEYWORD_VGRID, VGRID, ENVIRONMENTS
from .logging import create_localrunner_logfile, write_to_log
from .fileio import write_dir_pattern, write_dir_recipe, make_dir, \
    read_dir_recipe, read_dir_pattern, write_notebook, write_yaml, read_yaml, \
    delete_dir_pattern, delete_dir_recipe
from .meow import get_parameter_sweep_values, is_valid_pattern_object, Pattern
from .validation import valid_dir_path, check_input, is_valid_recipe_dict, \
    is_valid_local_environment

_trigger_event = '_trigger_event'

RUNNER_DATA = '.workflow_runner_data'
OUTPUT_DATA = 'job_output'

JOB_DIR = '.workflow_processing'

OP_CREATE = 'Created'
OP_MODIFIED = 'Modified'
OP_DELETED = 'Removed'
OP_EVENT = 'Event'

LOGGER = 'logger'
RULES = 'rules'
RULE_ID = 'id'
RULE_PATH = 'path'
RULE_PATTERN = 'pattern'
RULE_RECIPE = 'recipe'
ADMIN = 'admin'
MONITOR = 'monitor'
WORKERS = 'workers'
JOBS = 'jobs'
QUEUE = 'queue'
DATA_DIR = 'data_dir'
JOBS_DIR = 'jobs_dir'
OUTPUT_DIR = 'output_dir'
RETRO_ACTIVE = 'retro'
PRINT = 'print'

QUEUED = 'queued'
RUNNING = 'running'
FAILED = 'failed'
DONE = 'done'

JOB_ID = 'id'
JOB_PATTERN = 'pattern'
JOB_RECIPE = 'recipe'
JOB_RULE = 'rule'
JOB_PATH = 'path'
JOB_STATUS = 'status'
JOB_CREATE_TIME = 'create'
JOB_START_TIME = 'start'
JOB_END_TIME = 'end'
JOB_ERROR = 'error'
JOB_REQUIREMENTS = 'requirements'

META_FILE = 'job.yml'
BASE_FILE = 'base.ipynb'
PARAMS_FILE = 'params.yml'
JOB_FILE = 'job.ipynb'
RESULT_FILE = 'result.ipynb'


def get_runner_patterns(runner_data):
    return os.path.join(runner_data, PATTERNS)


def get_runner_recipes(runner_data):
    return os.path.join(runner_data, RECIPES)


def generate_id(length=16):
    """
    Generates a random id by using randomly generated alphanumeric strings.
    Uniqueness is not guaranteed, but is a reasonable assumption.

    :param length: (int) [optional] The length of the id to be generated.
    Default is 16

    :return: (str) A random collection of alphanumeric characters.
    """
    charset = CHAR_UPPERCASE + CHAR_LOWERCASE + CHAR_NUMERIC
    return ''.join(SystemRandom().choice(charset) for _ in range(length))


def replace_keywords(old_dict, job_id, src_path, vgrid):
    """
    Replaces MiG trigger keywords with with actual values.

    :param old_dict: (dict) A values dict potentially containing MiG keywords.

    :param job_id: (str) The appropriate job ID, corresponding to old_dict.

    :param src_path: (str) The triggering path for the event generating the
    job_id job

    :param vgrid: (str) The vgrid to replace

    :return: (dict) A dict corresponding to old_dict, with the MiG keywords
    replaced with appropriate values.
    """
    new_dict = {}

    filename = os.path.basename(src_path)
    dirname = os.path.dirname(src_path)
    relpath = os.path.relpath(src_path, vgrid)
    reldirname = os.path.dirname(relpath)
    (prefix, extension) = os.path.splitext(filename)

    for var, val in old_dict.items():
        if isinstance(val, str):
            val = val.replace(KEYWORD_PATH, src_path)
            val = val.replace(KEYWORD_REL_PATH, relpath)
            val = val.replace(KEYWORD_DIR, dirname)
            val = val.replace(KEYWORD_REL_DIR, reldirname)
            val = val.replace(KEYWORD_FILENAME, filename)
            val = val.replace(KEYWORD_PREFIX, prefix)
            val = val.replace(KEYWORD_VGRID, vgrid)
            val = val.replace(KEYWORD_EXTENSION, extension)
            val = val.replace(KEYWORD_JOB, job_id)

            new_dict[var] = val
        else:
            new_dict[var] = val

    return new_dict


def make_fake_event(path, state, is_directory=False):
    """Create a fake state change event for path. Looks up path to see if the
    change is a directory or file.
    """

    file_map = {'modified': FileModifiedEvent,
                'created': FileCreatedEvent,
                'deleted': FileDeletedEvent}
    dir_map = {'modified': DirModifiedEvent,
               'created': DirCreatedEvent, 'deleted': DirDeletedEvent}
    if is_directory or os.path.isdir(path):
        fake = dir_map[state](path)
    else:
        fake = file_map[state](path)

    # mark it a trigger event
    setattr(fake, _trigger_event, True)
    return fake


def administrator(
        from_user, to_user, from_state, from_file, to_queue, from_queue,
        to_worker_writers, from_worker_readers, to_logger, vgrid, job_data,
        meow_data, retro_active, workers_start):

    def add_pattern(pattern):
        op = OP_CREATE
        if pattern.name in patterns:
            if patterns[pattern.name] == pattern:
                return
            else:
                remove_pattern(pattern.name)
                op = OP_MODIFIED
        patterns[pattern.name] = pattern
        identify_rules(new_pattern=pattern)
        to_logger.send(
            (
                'add_pattern',
                '%s pattern %s' % (op, pattern)
            )
        )

    def add_recipe(recipe):
        op = OP_CREATE
        if recipe[NAME] in recipes:
            if recipes[recipe[NAME]] == recipe:
                return
            else:
                remove_recipe(recipe[NAME])
                op = OP_MODIFIED
        recipes[recipe[NAME]] = recipe
        identify_rules(new_recipe=recipe)
        to_logger.send(
            (
                'add_recipe',
                '%s recipe %s from source %s'
                % (op, recipe[NAME], recipe[SOURCE])
            )
        )

    def remove_pattern(pattern_name):
        if pattern_name in patterns:
            patterns.pop(pattern_name)
            remove_rules(deleted_pattern_name=pattern_name)
            to_logger.send(
                (
                    'remove_pattern',
                    'Removed pattern %s' % pattern_name
                )
            )
        else:
            to_logger.send(
                (
                    'remove_pattern',
                    'Pattern %s was not present in the pattern list to be '
                    'removed' % pattern_name
                )
            )

    def remove_recipe(recipe_name):
        if recipe_name in recipes:
            recipes.pop(recipe_name)
            remove_rules(deleted_recipe_name=recipe_name)
            to_logger.send(
                (
                    'remove_recipe',
                    'Removed recipe %s' % recipe_name
                )
            )
        else:
            to_logger.send(
                (
                    'remove_recipe',
                    'Recipe %s was not present in the recipe list to be '
                    'removed' % recipe_name
                )
            )

    def create_new_rule(pattern_name, recipe_name, path):
        rule = {
            RULE_ID: generate_id(),
            RULE_PATTERN: pattern_name,
            RULE_RECIPE: recipe_name,
            RULE_PATH: path
        }
        rules.append(rule)

        to_logger.send(
            (
                'create_new_rule',
                'Created rule for path: %s with id %s.'
                % (path, rule[RULE_ID])
            )
        )

        pattern = patterns[rule[RULE_PATTERN]]

        yaml_dict = {}
        for var, val in pattern.variables.items():
            yaml_dict[var] = val
        for var, val in pattern.outputs.items():
            yaml_dict[var] = val
        yaml_dict[pattern.trigger_file] = path

        if retro_active:
            testing_path = os.path.join(vgrid, path)

            globbed = glob.glob(testing_path)

            for globble in globbed:
                yaml_dict[pattern.trigger_file] = globble

                local_path = globble[globble.find(os.path.sep)+1:]

                if not pattern.sweep:
                    schedule_job(
                        rule,
                        local_path,
                        yaml_dict
                    )
                else:
                    for var, val in pattern.sweep.items():
                        values = get_parameter_sweep_values(val)
                        for value in values:
                            yaml_dict[var] = value
                            schedule_job(
                                rule,
                                local_path,
                                yaml_dict
                            )

    def identify_rules(new_pattern=None, new_recipe=None):
        if new_pattern:
            if len(new_pattern.recipes) > 1:
                to_logger.send(
                    (
                        'identify_rules-pattern',
                        'Rule creation aborted. Currently only supports one '
                        'recipe per pattern.'
                    )
                )
            recipe_name = new_pattern.recipes[0]
            if recipe_name in recipes:
                for input_path in new_pattern.trigger_paths:
                    create_new_rule(
                        new_pattern.name,
                        recipe_name,
                        input_path
                    )

        if new_recipe:
            for name, pattern in patterns.items():
                if len(pattern.recipes) > 1:
                    to_logger.send(
                        (
                            'identify_rules-recipe',
                            'Rule creation avoided for %s. Currently only '
                            'supports one recipe per pattern.' % name
                        )
                    )
                recipe_name = pattern.recipes[0]
                if recipe_name == new_recipe[NAME]:
                    for input_path in pattern.trigger_paths:
                        create_new_rule(
                            name,
                            recipe_name,
                            input_path
                        )

    def remove_rules(deleted_pattern_name=None, deleted_recipe_name=None):
        to_delete = []
        for rule in rules:
            if deleted_pattern_name:
                if rule[RULE_PATTERN] == deleted_pattern_name:
                    to_delete.append(rule)
            if deleted_recipe_name:
                if rule[RULE_RECIPE] == deleted_recipe_name:
                    to_delete.append(rule)
        for delete in to_delete:
            rules.remove(delete)
            to_logger.send(
                (
                    'remove_rules',
                    'Removing rule: %s.' % delete
                )
            )

    def schedule_job(rule, src_path, yaml_dict):
        """
        Schedules a new job in the workflow runner. This creates the
        appropriate job files in a shared directory, adds the job to the
        queue, and add it to the list of all jobs.

        :param rule: (dict) The rule causing this job to be scheduled.

        :param src_path: (str) The path which generated the triggering event.

        :param yaml_dict: (dict) Any variables to be applied.

        :return: No return.
        """
        recipe = recipes[rule[RULE_RECIPE]]

        environments = {}
        if ENVIRONMENTS in recipe \
                and 'local' in recipe[ENVIRONMENTS] \
                and is_valid_local_environment(recipe[ENVIRONMENTS]['local']):
            environments = recipe[ENVIRONMENTS]['local']

        job_dict = {
            JOB_ID: generate_id(),
            JOB_PATTERN: rule[RULE_PATTERN],
            JOB_RECIPE: rule[RULE_RECIPE],
            JOB_RULE: rule[RULE_ID],
            JOB_PATH: src_path,
            JOB_STATUS: QUEUED,
            JOB_CREATE_TIME: datetime.now(),
            JOB_REQUIREMENTS: environments
        }

        yaml_dict = replace_keywords(
            yaml_dict,
            job_dict[JOB_ID],
            src_path,
            vgrid
        )

        job_dir = os.path.join(job_data, job_dict[JOB_ID])
        make_dir(job_dir)

        meta_file = os.path.join(job_dir, META_FILE)
        write_yaml(job_dict, meta_file)

        base_file = os.path.join(job_dir, BASE_FILE)
        write_notebook(recipe[RECIPE], base_file)

        yaml_file = os.path.join(job_dir, PARAMS_FILE)
        write_yaml(yaml_dict, yaml_file)

        jobs.append(job_dict[JOB_ID])

        to_queue.send(job_dict[JOB_ID])

        to_logger.send(
            (
                'handle_event',
                'Scheduled new job %s from rule %s and pattern %s'
                % (job_dict[JOB_ID], rule[RULE_ID], rule[RULE_PATTERN])
            )
        )

    def handle_event(event):
        src_path = event.src_path
        event_type = event.event_type

        handle_path = src_path.replace(vgrid, '', 1)
        while handle_path.startswith(os.path.sep):
            handle_path = handle_path[1:]

        to_logger.send(
            (
                'handle_event',
                'Handling a %s event at %s' % (event_type, handle_path)
            )
        )

        for rule in rules:
            target_path = rule[RULE_PATH]
            recursive_regexp = fnmatch.translate(target_path)
            direct_regexp = recursive_regexp.replace('.*', '[^/]*')
            recursive_hit = re.match(recursive_regexp, handle_path)
            direct_hit = re.match(direct_regexp, handle_path)

            if direct_hit or recursive_hit:
                pattern = patterns[rule[RULE_PATTERN]]

                to_logger.send(
                    (
                        'run_handler',
                        'Starting new job for %s using rule %s'
                        % (src_path, rule)
                    )
                )

                yaml_dict = {}
                for var, val in pattern.variables.items():
                    yaml_dict[var] = val
                for var, val in pattern.outputs.items():
                    yaml_dict[var] = val
                yaml_dict[pattern.trigger_file] = src_path

                if not pattern.sweep:
                    schedule_job(
                        rule,
                        src_path,
                        yaml_dict
                    )
                else:
                    for var, val in pattern.sweep.items():
                        values = get_parameter_sweep_values(val)
                        for value in values:
                            yaml_dict[var] = value
                            schedule_job(
                                rule,
                                src_path,
                                yaml_dict
                            )

    def start_workers():
        for to_worker in to_worker_writers:
            to_worker.send('start')

        return True

    def stop_workers():
        for to_worker in to_worker_writers:
            to_worker.send('stop')

        return True

    def check_running_status():
        running, total = get_running_status()
        if running == total:
            return True, 'All workers are running. '
        else:
            return False, '%d workers are not running. ' % (total - running)

    def get_running_status():
        total = len(to_worker_writers)
        running = 0

        for i in range(len(to_worker_writers)):
            to_worker = to_worker_writers[i]
            to_worker.send('check')

            from_worker = from_worker_readers[i]
            status = from_worker.recv()
            if status == 'running':
                running += 1
        return running, total

    def stop_runner(clear_jobs=False):
        stop_workers()

        if os.path.exists(meow_data) \
                and os.path.isdir(meow_data) \
                and meow_data == RUNNER_DATA:
            shutil.rmtree(meow_data)

        if clear_jobs and os.path.exists(job_data):
            for job in jobs:
                job_dir = os.path.join(job_data, job)
                if os.path.exists(job_dir):
                    shutil.rmtree(job_dir)
            if len(os.listdir(job_data)) == 0:
                os.rmdir(job_data)
        return True

    def get_all_jobs():
        jobs_queue = copy.deepcopy(jobs)
        return jobs_queue

    def get_queued_jobs():
        to_queue.send('get_queue')
        queue = from_queue.recv()
        return queue

    def get_all_input_paths():
        input_paths = []
        for rule in rules:
            input_paths.append(rule[RULE_PATH])
        return input_paths

    def check_status():
        queued_jobs = len(get_queued_jobs())
        all_jobs = len(get_all_jobs())
        input_paths = get_all_input_paths()

        status = "[%s/%s] %s" % (queued_jobs, all_jobs, input_paths)

        return status

    def add_pattern_dir(pattern):
        status, msg = is_valid_pattern_object(pattern)
        if not status:
            to_logger.send(
                (
                    'add_pattern',
                    'Could not add pattern. %s' % msg
                )
            )

            return False
        else:
            write_dir_pattern(pattern, directory=meow_data)

            to_logger.send(
                (
                    'add_pattern',
                    'Added pattern: %s' % msg
                )
            )

            return True

    def modify_pattern_dir(pattern):
        status, msg = is_valid_pattern_object(pattern)
        if not status:
            to_logger.send(
                (
                    'modify_pattern',
                    'Could not modify pattern. %s' % msg
                )
            )

            return False
        else:
            write_dir_pattern(pattern, directory=meow_data)

            to_logger.send(
                (
                    'modify_pattern',
                    'Modified pattern: %s' % msg
                )
            )

            return True

    def remove_pattern_dir(pattern):
        if isinstance(pattern, str):
            name = pattern
        elif isinstance(pattern, Pattern):
            name = pattern.name
        else:
            to_logger.send(
                (
                    'remove_pattern',
                    'Invalid pattern parameter. Must be either Pattern '
                    'object, or a str name. '
                )
            )
            return False

        delete_dir_pattern(name, directory=meow_data)
        to_logger.send(
            (
                'remove_pattern',
                'Removed pattern: %s' % name
            )
        )

        return True

    def add_recipe_dir(recipe):
        status, msg = is_valid_recipe_dict(recipe)
        if not status:
            to_logger.send(
                (
                    'add_recipe',
                    'Could not add recipe. %s' % msg
                )
            )
            return False
        else:
            write_dir_recipe(recipe, directory=meow_data)

            to_logger.send(
                (
                    'add_recipe',
                    'Added user recipe: %s' % msg
                )
            )
            return True

    def modify_recipe_dir(recipe):
        status, msg = is_valid_recipe_dict(recipe)
        if not status:
            to_logger.send(
                (
                    'modify_recipe',
                    'Could not modify recipe. %s' % msg
                )
            )
            return False
        else:
            write_dir_recipe(recipe, directory=meow_data)

            to_logger.send(
                (
                    'modify_recipe',
                    'Modified recipe: %s' % msg
                )
            )
            return True

    def remove_recipe_dir(recipe):
        if isinstance(recipe, str):
            name = recipe
        elif isinstance(recipe, dict):
            name = recipe[NAME]
        else:
            to_logger.send(
                (
                    'remove_recipe',
                    'Invalid recipe parameter. Must be either recipe dict, '
                    'or a str name. '
                )
            )
            return False

        delete_dir_recipe(name, directory=meow_data)
        to_logger.send(
            (
                'remove_recipe',
                'Removed recipe: %s' % name
            )
        )
        return True

    def check_recipes():
        recipes_dict = copy.deepcopy(recipes)
        return recipes_dict

    def check_patterns():
        patterns_dict = copy.deepcopy(patterns)
        return patterns_dict

    def check_rules():
        rules_list = copy.deepcopy(rules)
        return rules_list

    def check_jobs():
        jobs_list = copy.deepcopy(jobs)
        return jobs_list

    def check_queue():
        return get_queued_jobs()

    # Start of administrator

    patterns = {}
    recipes = {}
    rules = []
    jobs = []

    if workers_start:
        start_workers()

    while True:
        ready = wait([
            from_state,
            from_user,
            from_file
        ])

        if from_state in ready:
            input_message = from_state.recv()
            operation = input_message['operation']
            if 'pattern' in input_message:
                if operation == OP_CREATE:
                    add_pattern(input_message['pattern'])
                elif operation == OP_DELETED:
                    remove_pattern(input_message['pattern'])
            elif 'recipe' in input_message:
                if operation == OP_CREATE:
                    add_recipe(input_message['recipe'])
                elif operation == OP_DELETED:
                    remove_recipe(input_message['recipe'])

        elif from_user in ready:
            input_message = from_user.recv()
            operation = input_message[0]
            args = input_message[1]

            if operation == 'start_workers':
                result = start_workers()
                to_user.send(result)

            elif operation == 'stop_workers':
                result = stop_workers()
                to_user.send(result)

            elif operation == 'check_running_status':
                result = check_running_status()
                to_user.send(result)

            elif operation == 'get_running_status':
                result = get_running_status()
                to_user.send(result)

            elif operation == 'stop_runner':
                result = stop_runner(clear_jobs=args)
                to_user.send(result)

            elif operation == 'get_all_jobs':
                result = get_all_jobs()
                to_user.send(result)

            elif operation == 'get_queued_jobs':
                result = get_queued_jobs()
                to_user.send(result)

            elif operation == 'get_all_input_paths':
                result = get_all_input_paths()
                to_user.send(result)

            elif operation == 'check_status':
                result = check_status()
                to_user.send(result)

            elif operation == 'add_pattern':
                result = add_pattern_dir(args)
                to_user.send(result)

            elif operation == 'modify_pattern':
                result = modify_pattern_dir(args)
                to_user.send(result)

            elif operation == 'remove_pattern':
                result = remove_pattern_dir(args)
                to_user.send(result)

            elif operation == 'add_recipe':
                result = add_recipe_dir(args)
                to_user.send(result)

            elif operation == 'modify_recipe':
                result = modify_recipe_dir(args)
                to_user.send(result)

            elif operation == 'remove_recipe':
                result = remove_recipe_dir(args)
                to_user.send(result)

            elif operation == 'check_recipes':
                result = check_recipes()
                to_user.send(result)

            elif operation == 'check_patterns':
                result = check_patterns()
                to_user.send(result)

            elif operation == 'check_rules':
                result = check_rules()
                to_user.send(result)

            elif operation == 'check_jobs':
                result = check_jobs()
                to_user.send(result)

            elif operation == 'check_queue':
                result = check_queue()
                to_user.send(result)

            elif operation == 'kill':
                to_user.send('dead')
                return

            else:
                raise Exception('Unknown message format: %s' % input_message)

        elif from_file in ready:
            input_message = from_file.recv()
            handle_event(input_message)


def job_queue(from_admin, to_admin, from_worker_readers, to_worker_writers,
              to_logger, job_home):
    queue = []
    worker_module_lists = []

    all_inputs = [from_admin]

    for channel_reader in from_worker_readers:
        all_inputs.append(channel_reader)
        worker_module_lists.append([])

    while True:
        ready = wait(all_inputs)

        if from_admin in ready:
            input_message = from_admin.recv()
            if input_message == 'get_queue':
                current_queue = copy.deepcopy(queue)
                to_admin.send(current_queue)

            elif input_message == 'kill':
                to_admin.send('dead')
                return

            # submitting new job
            else:
                print('appended to queue')
                queue.append(input_message)

        # Is from worker
        else:
            for i in range(len(from_worker_readers)):
                if from_worker_readers[i] in ready:
                    input_message = from_worker_readers[i].recv()
                    to_worker = to_worker_writers[i]
                    # Is module list
                    if isinstance(input_message, list):
                        worker_module_lists[i] = input_message
                    if input_message == 'request':
                        print('got request')
                        assigned_job = None
                        for job_id in queue:
                            print('considering %s' % job_id)
                            job_dir = os.path.join(job_home, job_id)
                            meta_path = os.path.join(job_dir, META_FILE)
                            job_data = read_yaml(meta_path)

                            requirements = job_data[JOB_REQUIREMENTS]
                            missing_requirement = False
                            for requirement in requirements['dependencies']:
                                print(('has requirement %s' % requirement))
                                if requirement not in worker_module_lists[i]:
                                    print('missing %s' % requirement)
                                    missing_requirement = True
                            if not missing_requirement:
                                assigned_job = job_id
                                break
                            else:
                                to_logger.send(
                                    (
                                        'queue request',
                                        "Could not assign job %s to worker "
                                        "%s as missing one or more "
                                        "requirement from %s."
                                        % (job_id, i, requirements)
                                    )
                                )
                        if assigned_job:
                            queue.remove(assigned_job)

                            to_logger.send(
                                (
                                    'queue request',
                                    "Assigning job %s" % assigned_job
                                )
                            )
                        print('sending job %s' % assigned_job)
                        to_worker.send(assigned_job)


def job_processor(from_timer, to_timer, from_admin, to_admin, to_queue,
                  from_queue, to_logger, processor_id, job_home, output_data):
    state = 'stopped'

    to_timer.send('sleep')

    module_list = [p.project_name for p in pkg_resources.working_set]
    to_queue.send(module_list)

    while True:
        ready = wait([from_admin, from_timer])

        if from_admin in ready:
            input_message = from_admin.recv()

            if input_message == 'start':
                state = 'running'

            elif input_message == 'check':
                to_admin.send(state)

            elif input_message == 'stop':
                state = 'stopped'

            elif input_message == 'kill':
                to_timer.send('kill')
                to_admin.send('dead')
                return

        elif from_timer in ready:
            input_message = from_timer.recv()

            if input_message == 'done' and state == 'running':
                to_queue.send('request')

                job_id = from_queue.recv()

                if job_id:
                    job_dir = os.path.join(job_home, job_id)
                    meta_path = os.path.join(job_dir, META_FILE)
                    base_path = os.path.join(job_dir, BASE_FILE)
                    param_path = os.path.join(job_dir, PARAMS_FILE)
                    job_path = os.path.join(job_dir, JOB_FILE)
                    result_path = os.path.join(job_dir, RESULT_FILE)

                    job_data = read_yaml(meta_path)

                    job_data[JOB_STATUS] = RUNNING
                    job_data[JOB_START_TIME] = datetime.now()

                    write_yaml(job_data, meta_path)

                    to_logger.send(
                        (
                            'worker %s' % processor_id,
                            "Found job %s" % job_data[JOB_ID]
                        )
                    )

                    error = False
                    cmd = 'notebook_parameterizer ' \
                          + base_path + ' ' \
                          + param_path + ' ' \
                          + '-o ' + job_path
                    try:
                        sub = subprocess.Popen(cmd,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE,
                                               shell=True)
                        # TODO: implement a timeout (max simulation time)
                        (stdout, stderr) = sub.communicate()

                    except Exception as ex:
                        error = ex

                    if not os.path.exists(job_path) or error:
                        job_data[JOB_STATUS] = FAILED
                        job_data[JOB_END_TIME] = datetime.now()
                        msg = 'Job file %s was not created successfully' \
                              % job_id
                        if error:
                            msg += '. %s' % error
                        job_data[JOB_ERROR] = msg
                        write_yaml(job_data, meta_path)
                        to_logger.send(
                            (
                                'worker %s' % processor_id,
                                "Job worker encountered an error. %s" % msg
                            )
                        )
                        continue

                    cmd = 'papermill ' \
                          + job_path + ' ' \
                          + result_path
                    try:
                        sub = subprocess.Popen(cmd,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE,
                                               shell=True)
                        # TODO: implement a timeout (max simulation time)
                        (stdout, stderr) = sub.communicate()
                    except Exception as ex:
                        error = ex

                    if not os.path.exists(result_path) or error:
                        job_data[JOB_STATUS] = FAILED
                        job_data[JOB_END_TIME] = datetime.now()
                        msg = 'Result file %s was not created successfully'
                        if error:
                            msg += '. %s' % error
                        job_data[JOB_ERROR] = msg
                        write_yaml(job_data, meta_path)
                        to_logger.send(
                            (
                                'worker %s' % processor_id,
                                "Job worker encountered an error. %s" % msg
                            )
                        )
                        continue

                    job_data[JOB_STATUS] = DONE
                    job_data[JOB_END_TIME] = datetime.now()
                    write_yaml(job_data, meta_path)

                    job_output_dir = os.path.join(output_data, job_id)

                    shutil.copytree(job_dir, job_output_dir)

                    to_logger.send(
                        (
                            'worker %s' % processor_id,
                            "Completed job %s" % job_data[JOB_ID]
                        )
                    )

                else:
                    to_logger.send(
                        (
                            'worker %s' % processor_id,
                            "Worker %s found no job in queue" % processor_id
                        )
                    )

            to_timer.send('sleep')


def worker_timer(from_worker, to_worker, worker_id, wait_time=10):
    while True:
        msg = from_worker.recv()

        if msg == 'sleep':
            time.sleep(wait_time + (worker_id % wait_time))

            to_worker.send('done')
        elif msg == 'kill':
            to_worker.send('dead')
            return


def logger(all_input_channel_readers, print_logging=True, file_logging=False):
    runner_log_file = create_localrunner_logfile(debug_mode=file_logging)

    while True:
        ready = wait(all_input_channel_readers)

        for reader in all_input_channel_readers:
            if reader in ready:
                input_message = reader.recv()

                write_to_log(
                    runner_log_file,
                    input_message[0],
                    input_message[1]
                )

                if print_logging:
                    print(input_message[1])

                continue


class WorkflowRunner:
    def __init__(self, path, workers, patterns=None, recipes=None,
                 meow_data=RUNNER_DATA, job_data=JOB_DIR,
                 output_data=OUTPUT_DATA, daemon=False, reuse_vgrid=True,
                 start_workers=True, retro_active_jobs=True,
                 print_logging=True, file_logging=False):

        valid_dir_path(path, 'path')
        check_input(workers, int, 'workers', or_none=True)
        check_input(patterns, dict, PATTERNS, or_none=True)
        check_input(recipes, dict, RECIPES, or_none=True)
        valid_dir_path(meow_data, 'meow_data')
        valid_dir_path(job_data, 'job_data')
        valid_dir_path(output_data, 'output_data')
        check_input(daemon, bool, 'daemon')
        check_input(reuse_vgrid, bool, 'reuse_vgrid')
        check_input(start_workers, bool, 'start_workers')
        check_input(retro_active_jobs, bool, 'retro_active_jobs')
        check_input(print_logging, bool, 'print_logging')
        check_input(file_logging, bool, 'file_logging')

        make_dir(path, can_exist=reuse_vgrid)
        make_dir(job_data)
        if meow_data == RUNNER_DATA:
            make_dir(meow_data, ensure_clean=True)
        else:
            make_dir(meow_data)
        make_dir(output_data)
        make_dir(get_runner_patterns(meow_data), ensure_clean=True)
        make_dir(get_runner_recipes(meow_data), ensure_clean=True)

        user_to_admin_reader, user_to_admin_writer = Pipe(duplex=False)
        admin_to_user_reader, admin_to_user_writer = Pipe(duplex=False)
        user_to_logger_reader, user_to_logger_writer = Pipe(duplex=False)

        self.user_to_admin = user_to_admin_writer
        self.admin_to_user = admin_to_user_reader

        state_to_admin_reader, state_to_admin_writer = Pipe(duplex=False)
        state_to_logger_reader, state_to_logger_writer = Pipe(duplex=False)

        file_to_admin_reader, file_to_admin_writer = Pipe(duplex=False)
        file_to_logger_reader, file_to_logger_writer = Pipe(duplex=False)

        admin_to_queue_reader, admin_to_queue_writer = Pipe(duplex=False)
        queue_to_admin_reader, queue_to_admin_writer = Pipe(duplex=False)
        admin_to_logger_reader, admin_to_logger_writer = Pipe(duplex=False)

        queue_to_logger_reader, queue_to_logger_writer = Pipe(duplex=False)

        all_logger_inputs = [
            user_to_logger_reader,
            state_to_logger_reader,
            file_to_logger_reader,
            admin_to_logger_reader,
            queue_to_logger_reader,
        ]

        workers_and_timers_list = []
        admin_to_workers = []
        worker_to_admins = []
        worker_to_queues = []
        queue_to_workers = []
        for processor_id in range(0, workers):
            worker_to_timer_reader, worker_to_timer_writer = Pipe(duplex=False)
            timer_to_worker_reader, timer_to_worker_writer = Pipe(duplex=False)
            admin_to_worker_reader, admin_to_worker_writer = Pipe(duplex=False)
            worker_to_admin_reader, worker_to_admin_writer = Pipe(duplex=False)
            worker_to_queue_reader, worker_to_queue_writer = Pipe(duplex=False)
            queue_to_worker_reader, queue_to_worker_writer = Pipe(duplex=False)
            worker_to_logger_reader, worker_to_logger_writer = Pipe(duplex=False)

            worker = Process(
                target=job_processor,
                args=(
                    timer_to_worker_reader,
                    worker_to_timer_writer,
                    admin_to_worker_reader,
                    worker_to_admin_writer,
                    worker_to_queue_writer,
                    queue_to_worker_reader,
                    worker_to_logger_writer,
                    processor_id,
                    job_data,
                    output_data
                )
            )

            timer = Process(
                target=worker_timer,
                args=(
                    worker_to_timer_reader,
                    timer_to_worker_writer,
                    processor_id
                )
            )

            workers_and_timers_list.append(worker)
            workers_and_timers_list.append(timer)
            admin_to_workers.append(admin_to_worker_writer)
            worker_to_admins.append(worker_to_admin_reader)
            worker_to_queues.append(worker_to_queue_reader)
            queue_to_workers.append(queue_to_worker_writer)
            all_logger_inputs.append(worker_to_logger_reader)

        administrator_process = Process(
            target=administrator,
            args=(
                user_to_admin_reader,
                admin_to_user_writer,
                state_to_admin_reader,
                file_to_admin_reader,
                admin_to_queue_writer,
                queue_to_admin_reader,
                admin_to_workers,
                worker_to_admins,
                admin_to_logger_writer,
                path,
                job_data,
                meow_data,
                retro_active_jobs,
                start_workers
            )
        )

        job_queue_process = Process(
            target=job_queue,
            args=(
                admin_to_queue_reader,
                queue_to_admin_writer,
                worker_to_queues,
                queue_to_workers,
                queue_to_logger_writer,
                job_data
            )
        )

        logger_process = Process(
            target=logger,
            args=(
                all_logger_inputs,
                print_logging,
                file_logging
            )
        )

        process_list = [
            logger_process,
            job_queue_process,
            administrator_process,
        ]

        self.process_list = process_list
        for worker_or_timer in workers_and_timers_list:
            process_list.append(worker_or_timer)

        # Start all non-monitoring processes
        self.run()

        state_monitor = LocalWorkflowStateMonitor(
            state_to_admin_writer, state_to_logger_writer, meow_data)
        state_monitor_process = Observer()
        state_monitor_process.schedule(
            state_monitor,
            meow_data,
            recursive=True
        )

        file_monitor = LocalWorkflowFileMonitor(
            file_to_admin_writer, file_to_logger_writer)
        file_monitor_process = Observer()
        file_monitor_process.schedule(
            file_monitor,
            path,
            recursive=True
        )

        self.process_list.append(state_monitor_process)
        self.process_list.append(file_monitor_process)

        state_monitor_process.start()
        file_monitor_process.start()

        admin_to_logger_writer.send(
            (
                'WorkflowRunner._init',
                 'Started WorkflowRunner processes. '
            )
        )

        if patterns:
            for name, pattern in patterns.items():
                admin_to_logger_writer.send(
                    (
                        'WorkflowRunner._init',
                        "Adding pattern %s. " % name
                    )
                )

                write_dir_pattern(pattern, directory=meow_data)

        admin_to_logger_writer.send(
            (
                'WorkflowRunner._init',
                'Added all predefined patterns. '
            )
        )

        if recipes:
            for name, recipe in recipes.items():
                admin_to_logger_writer.send(
                    (
                        'WorkflowRunner._init',
                        "Adding recipe %s. " % name
                    )
                )

                write_dir_recipe(recipe, directory=meow_data)

        admin_to_logger_writer.send(
            (
                'WorkflowRunner._init',
                 'Added all predefined recipes. '
            )
        )

        if not daemon:
            try:
                while True:
                    time.sleep(1)
            except KeyboardInterrupt:
                self.stop_runner()

    def run(self):
        for my_process in self.process_list:
            my_process.start()

    def join(self):
        for my_process in self.process_list:
            my_process.join()

    def stop(self):
        for my_process in self.process_list:
            if hasattr(my_process, 'terminate'):
                my_process.terminate()

    def start_workers(self):
        self.user_to_admin.send(
            (
                'start_workers',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def stop_workers(self):
        self.user_to_admin.send(
            (
                'stop_workers',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def check_running_status(self):
        self.user_to_admin.send(
            (
                'check_running_status',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def get_running_status(self):
        self.user_to_admin.send(
            (
                'get_running_status',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def stop_runner(self, clear_jobs=False):
        self.user_to_admin.send(
            (
                'stop_runner',
                clear_jobs
            )
        )
        result = self.admin_to_user.recv()
        self.stop()
        return result

    def get_all_jobs(self):
        self.user_to_admin.send(
            (
                'get_all_jobs',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def get_queued_jobs(self):
        self.user_to_admin.send(
            (
                'get_queued_jobs',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def get_all_input_paths(self):
        self.user_to_admin.send(
            (
                'get_all_input_paths',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def check_status(self):
        self.user_to_admin.send(
            (
                'check_status',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def add_pattern(self, pattern):
        self.user_to_admin.send(
            (
                'add_pattern',
                pattern
            )
        )
        result = self.admin_to_user.recv()
        return result

    def modify_pattern(self, pattern):
        self.user_to_admin.send(
            (
                'modify_pattern',
                pattern
            )
        )
        result = self.admin_to_user.recv()
        return result

    def remove_pattern(self, pattern):
        self.user_to_admin.send(
            (
                'remove_pattern',
                pattern
            )
        )
        result = self.admin_to_user.recv()
        return result

    def add_recipe(self, recipe):
        self.user_to_admin.send(
            (
                'add_recipe',
                recipe
            )
        )
        result = self.admin_to_user.recv()
        return result

    def modify_recipe(self, recipe):
        self.user_to_admin.send(
            (
                'modify_recipe',
                recipe
            )
        )
        result = self.admin_to_user.recv()
        return result

    def remove_recipe(self, recipe):
        self.user_to_admin.send(
            (
                'remove_recipe',
                recipe
            )
        )
        result = self.admin_to_user.recv()
        return result

    def check_recipes(self):
        self.user_to_admin.send(
            (
                'check_recipes',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def check_patterns(self):
        self.user_to_admin.send(
            (
                'check_patterns',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def check_rules(self):
        self.user_to_admin.send(
            (
                'check_rules',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def check_jobs(self):
        self.user_to_admin.send(
            (
                'check_jobs',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result

    def check_queue(self):
        self.user_to_admin.send(
            (
                'check_queue',
                None
            )
        )
        result = self.admin_to_user.recv()
        return result


class LocalWorkflowStateMonitor(PatternMatchingEventHandler):
    """
    Event handler to monitor pattern and recipe changes.
    """

    def __init__(
            self, to_admin, to_logger,  meow_data, patterns=None,
            ignore_patterns=None, ignore_directories=False,
            case_sensitive=False):
        """Constructor"""

        PatternMatchingEventHandler.__init__(
            self,
            patterns,
            ignore_patterns,
            ignore_directories,
            case_sensitive
        )
        self.to_logger = to_logger
        self.to_admin = to_admin
        self.meow_data = meow_data
        self.state_patterns = {}
        self.state_recipes = {}

        self.to_logger.send(
            (
                'LocalWorkflowStateMonitor._init',
                'Setting up new state monitor'
            )
        )

        # Check pre-existing Patterns and Recipes
        runner_patterns = get_runner_patterns(self.meow_data)
        for file_path in os.listdir(runner_patterns):
            try:
                pattern = read_dir_pattern(
                    file_path,
                    directory=self.meow_data
                )

                self.add_pattern(pattern)
            except Exception as exc:
                self.to_logger.send(
                    ('LocalWorkflowStateMonitor._init', str(exc))
                )

        runner_recipes = get_runner_recipes(self.meow_data)
        for file_path in os.listdir(runner_recipes):
            try:
                recipe = read_dir_recipe(
                    file_path,
                    directory=self.meow_data
                )
                # self.to_logger.send(
                #     (
                #         'LocalWorkflowStateMonitor._init',
                #         'Identified pre-existing recipe: %s' % pattern.name
                #     )
                # )
                self.add_recipe(recipe)
            except Exception as exc:
                self.to_logger.send(
                    ('LocalWorkflowStateMonitor._init', str(exc))
                )
        self.to_logger.send(
            (
                'LocalWorkflowStateMonitor._init',
                'Finished identifying pre-existing state'
            )
        )

    def update_rules(self, event):
        """Handle all rule updates"""

        if event.is_directory:
            return

        self.to_logger.send(
            (
                'update_rules',
                "Handling %s rule update at %s"
                % (event.event_type, event.src_path)
            )
        )

        src_path = event.src_path

        try:
            valid_dir_path(src_path, 'src_path')
        except ValueError as ve:
            self.to_logger.send(
                (
                    'update_rules',
                    "Ignoring events at %s. %s" % (src_path, ve)
                )
            )
            return

        event_type = event.event_type
        file_type = ''
        file_path = ''
        try:
            runner_patterns = get_runner_patterns(self.meow_data)
            runner_recipes = get_runner_recipes(self.meow_data)
            if runner_patterns in src_path:
                file_path = src_path[
                            src_path.find(runner_patterns)
                            + len(runner_patterns)+1:]
                file_type = PATTERNS
            elif runner_recipes in src_path:
                file_path = src_path[
                            src_path.find(runner_recipes)
                            + len(runner_recipes)+1:]
                file_type = RECIPES
        except Exception as exc:
            self.to_logger.send(
                (
                    'update_rules-pattern',
                    'Cannot process event at %s due to error: %s'
                    % (src_path, exc)
                )
            )
            return
        if os.path.sep in file_path:
            self.to_logger.send(
                (
                    'update_rules-pattern',
                    'Cannot process nested event at %s' % src_path
                )
            )
            return

        if event_type in ['created', 'modified']:
            if file_type == PATTERNS:
                try:
                    pattern = read_dir_pattern(
                        file_path,
                        directory=self.meow_data
                    )
                except Exception as exc:
                    self.to_logger.send(
                        (
                            'update_rules-pattern',
                            str(exc)
                        )
                    )
                    return
                self.add_pattern(pattern)
            elif file_type == RECIPES:
                try:
                    recipe = read_dir_recipe(
                        file_path,
                        directory=self.meow_data
                    )
                except Exception as exc:
                    self.to_logger.send(
                        (
                            'update_rules-recipe',
                            str(exc)
                        )
                    )
                    return
                self.add_recipe(recipe)
        elif event_type == 'deleted':
            if file_type == PATTERNS:
                self.remove_pattern(file_path)
            elif file_type == RECIPES:
                self.remove_recipe(file_path)

    def on_modified(self, event):
        """Handle modified rule file"""

        self.update_rules(event)

    def on_created(self, event):
        """Handle new rule file"""

        self.update_rules(event)

    def on_deleted(self, event):
        """Handle deleted rule file"""

        self.update_rules(event)

    def add_pattern(self, pattern):
        op = OP_CREATE
        if pattern.name in self.state_patterns:
            if self.state_patterns[pattern.name] == pattern:
                return
            else:
                self.remove_pattern(pattern.name)
                op = OP_MODIFIED
        self.state_patterns[pattern.name] = pattern

        msg = {
            'operation': OP_CREATE,
            'pattern': pattern
        }
        self.to_admin.send(msg)

        self.to_logger.send(
            (
                'add_pattern',
                '%s pattern %s' % (op, pattern)
            )
        )

    def add_recipe(self, recipe):
        op = OP_CREATE
        if recipe[NAME] in self.state_recipes:
            if self.state_recipes[recipe[NAME]] == recipe:
                return
            else:
                self.remove_recipe(recipe[NAME])
                op = OP_MODIFIED
        self.state_recipes[recipe[NAME]] = recipe

        msg = {
            'operation': OP_CREATE,
            'recipe': recipe
        }
        self.to_admin.send(msg)

        self.to_logger.send(
            (
                'add_recipe',
                '%s recipe %s from source %s'
                % (op, recipe[NAME], recipe[SOURCE])
            )
        )

    def remove_pattern(self, pattern_name):
        if pattern_name in self.state_patterns:
            self.state_patterns.pop(pattern_name)

            msg = {
                'operation': OP_DELETED,
                'pattern': pattern_name
            }
            self.to_admin.send(msg)

            self.to_logger.send(
                (
                    'remove_pattern',
                    'Removed pattern %s' % pattern_name
                )
            )
        else:
            self.to_logger.send(
                (
                    'remove_pattern',
                    'Pattern %s was not present in the pattern list to be '
                    'removed' % pattern_name
                )
            )

    def remove_recipe(self, recipe_name):
        if recipe_name in self.state_recipes:
            self.state_recipes.pop(recipe_name)

            msg = {
                'operation': OP_DELETED,
                'recipe': recipe_name
            }
            self.to_admin.send(msg)

            self.to_logger.send(
                (
                    'remove_recipe',
                    'Removed recipe %s' % recipe_name
                )
            )
        else:
            self.to_logger.send(
                (
                    'remove_recipe',
                    'Recipe %s was not present in the recipe list to be '
                    'removed' % recipe_name
                )
            )


class LocalWorkflowFileMonitor(PatternMatchingEventHandler):
    """
    Event handler to schedule jobs according to file events.
    """

    def __init__(
            self, to_admin, to_logger, patterns=None,
            ignore_patterns=None, ignore_directories=False,
            case_sensitive=False):
        """Constructor"""

        PatternMatchingEventHandler.__init__(
            self,
            patterns,
            ignore_patterns,
            ignore_directories,
            case_sensitive
        )
        self.to_logger = to_logger
        self.to_admin = to_admin
        self.recent_jobs = {}
        self._recent_jobs_lock = threading.Lock()

        self.to_logger.send(
            (
                'LocalWorkflowRunner',
                'Setting up new file monitor'
            )
        )

    def __handle_trigger(self, event):
        pid = current_process().pid
        event_type = event.event_type
        src_path = event.src_path
        time_stamp = event.time_stamp

        # self.to_logger.send(
        #     (
        #         '__handle_trigger',
        #         'Running threaded handler at (%s) to handle %s event at %s at '
        #         '%s' % (pid, event_type, src_path, time_stamp)
        #     )
        # )

        # This will prevent some job spamming
        self._recent_jobs_lock.acquire()
        try:
            if src_path in self.recent_jobs:
                recent_timestamp = self.recent_jobs[src_path]
                difference = time_stamp - recent_timestamp

                if difference <= 1:
                    self.recent_jobs[src_path] = \
                        max(recent_timestamp, time_stamp)
                    # self.to_logger.send(
                    #     (
                    #         '__handle_trigger',
                    #         'Skipping due to recent hit'
                    #     )
                    # )
                    self._recent_jobs_lock.release()
                    return
                else:
                    self.recent_jobs[src_path] = time_stamp
            else:
                self.recent_jobs[src_path] = time_stamp
        except Exception as ex:
            self._recent_jobs_lock.release()
            raise Exception(ex)
        self._recent_jobs_lock.release()

        self.to_admin.send(event)

    def run_handler(self, event):
        waiting_for_threaded_resources = True
        while waiting_for_threaded_resources:
            try:
                worker = threading.Thread(
                    target=self.__handle_trigger,
                    args=[event])
                worker.daemon = True
                worker.start()
                waiting_for_threaded_resources = False
            except threading.ThreadError:
                time.sleep(1)

    def handle_event(self, event):
        if event.is_directory:
            return

        if event.event_type not in ['created', 'modified']:
            return

        event.time_stamp = time.time()

        self.run_handler(event)

    def on_modified(self, event):
        """Handle modified files"""

        self.handle_event(event)

    def on_created(self, event):
        """Handle created files"""

        self.handle_event(event)

    def on_deleted(self, event):
        """Handle deleted files"""

        self.handle_event(event)

    def on_moved(self, event):
        """Handle moved files"""

        fake = make_fake_event(
            event.dest_path,
            'created',
            event.is_directory
        )
        self.handle_event(fake)




# Only constants shared amongst multiple files should be declared here.
# Variables that are only used in one file should be declared there, unless it
# is clearer to group them here with appropriate variables.
# Variables may also need to be saved here if needed for testing.

DEFAULT_JSON_TIMEOUT = 60
DEFAULT_WORKFLOW_TITLE = 'workflow'
DEFAULT_MEOW_IMPORT_EXPORT_DIR = 'meow_directory'
DEFAULT_CWL_IMPORT_EXPORT_DIR = 'cwl_directory'

WORKFLOWS = 'workflows'
STEPS = 'steps'
SETTINGS = 'variables'

MEOW_NEW_PATTERN_BUTTON = 'meow_new_pattern_button'
MEOW_EDIT_PATTERN_BUTTON = 'meow_edit_pattern_button'
MEOW_NEW_RECIPE_BUTTON = 'meow_new_recipe_button'
MEOW_EDIT_RECIPE_BUTTON = 'meow_edit_recipe_button'
MEOW_IMPORT_CWL_BUTTON = 'meow_import_cwl_button'
MEOW_IMPORT_VGRID_BUTTON = 'meow_import_vgrid_button'
MEOW_EXPORT_VGRID_BUTTON = 'meow_export_vgrid_button'
MEOW_IMPORT_DIR_BUTTON = 'meow_import_dir_button'
MEOW_EXPORT_DIR_BUTTON = 'meow_export_dir_button'
MEOW_SAVE_SVG_BUTTON = 'meow_save_svg_button'

CWL_NEW_WORKFLOW_BUTTON = 'cwl_new_workflow_button'
CWL_EDIT_WORKFLOW_BUTTON = 'cwl_edit_workflow_button'
CWL_NEW_STEP_BUTTON = 'cwl_new_step_button'
CWL_EDIT_STEP_BUTTON = 'cwl_edit_step_button'
CWL_NEW_VARIABLES_BUTTON = 'cwl_new_variables_button'
CWL_EDIT_VARIABLES_BUTTON = 'cwl_edit_variables_button'
CWL_IMPORT_MEOW_BUTTON = 'cwl_import_meow_button'
CWL_IMPORT_DIR_BUTTON = 'cwl_import_dir_button'
CWL_EXPORT_DIR_BUTTON = 'cwl_export_dir_button'
CWL_SAVE_SVG_BUTTON = 'cwl_save_svg_button'

CWL_NAME = 'name'
CWL_CWL_VERSION = 'cwlVersion'
CWL_CLASS = 'class'
CWL_BASE_COMMAND = 'baseCommand'
CWL_INPUTS = 'inputs'
CWL_OUTPUTS = 'outputs'
CWL_ARGUMENTS = 'arguments'
CWL_REQUIREMENTS = 'requirements'
CWL_HINTS = 'hints'
CWL_STDOUT = 'stdout'
CWL_STEPS = 'steps'
CWL_VARIABLES = 'arguments'

CWL_OUTPUT_TYPE = 'type'
CWL_OUTPUT_BINDING = 'outputBinding'
CWL_OUTPUT_SOURCE = 'outputSource'
CWL_OUTPUT_GLOB = 'glob'

CWL_YAML_CLASS = 'class'
CWL_YAML_PATH = 'path'

CWL_WORKFLOW_RUN = 'run'
CWL_WORKFLOW_IN = 'in'
CWL_WORKFLOW_OUT = 'out'

DEFAULT_JOB_NAME = 'wf_job'

PATTERN_LIST = 'pattern_list'
RECIPE_LIST = 'recipe_list'

VGRID_WORKFLOWS_OBJECT = 'workflows'
VGRID_PATTERN_OBJECT_TYPE = 'workflowpattern'
VGRID_RECIPE_OBJECT_TYPE = 'workflowrecipe'
VGRID_ANY_OBJECT_TYPE = 'any'
VGRID_REPORT_OBJECT_TYPE = 'workflow_report'

VALID_WORKFLOW_TYPES = [
    VGRID_WORKFLOWS_OBJECT,
    VGRID_PATTERN_OBJECT_TYPE,
    VGRID_RECIPE_OBJECT_TYPE,
    VGRID_ANY_OBJECT_TYPE
]

VGRID_QUEUE_OBJECT_TYPE = 'queue'
VGRID_JOB_OBJECT_TYPE = 'job'
CANCEL_JOB = 'cancel_job'
RESUBMIT_JOB = 'resubmit_job'

VALID_JOB_TYPES = [
    VGRID_QUEUE_OBJECT_TYPE,
    VGRID_JOB_OBJECT_TYPE,
    CANCEL_JOB,
    RESUBMIT_JOB
]

VGRID_ERROR_TYPE = 'error_text'
VGRID_TEXT_TYPE = 'text'
VGRID_CREATE = 'create'
VGRID_READ = 'read'
VGRID_UPDATE = 'update'
VGRID_DELETE = 'delete'

VALID_OPERATIONS = [
    VGRID_CREATE,
    VGRID_READ,
    VGRID_UPDATE,
    VGRID_DELETE
]

OBJECT_TYPE = 'object_type'
PERSISTENCE_ID = 'persistence_id'
NAME = 'name'
INPUT_FILE = 'input_file'
TRIGGER_PATHS = 'input_paths'
OUTPUT = 'output'
RECIPE = 'recipe'
RECIPES = 'recipes'
VARIABLES = 'variables'
SWEEP = 'parameterize_over'
SWEEP_START = 'start'
SWEEP_STOP = 'stop'
SWEEP_JUMP = 'increment'
ENVIRONMENTS = 'environments'
VGRID = 'vgrid'
SOURCE = 'source'
PATTERNS = 'patterns'
TRIGGER_RECIPES = 'trigger_recipes'
TASK_FILE = 'task_file'

CWL_CLASS_COMMAND_LINE_TOOL = 'CommandLineTool'
CWL_CLASS_WORKFLOW = 'Workflow'

PATTERN_NAME = 'Pattern'
RECIPE_NAME = 'Recipe'
WORKFLOW_NAME = 'Workflow'
STEP_NAME = 'Step'
VARIABLES_NAME = 'Arguments'

PLACEHOLDER = 'PLACEHOLDER'

VALID_PATTERN_MIN = {
    NAME: str,
    INPUT_FILE: str,
    TRIGGER_PATHS: list,
    TRIGGER_RECIPES: dict
}

VALID_PATTERN_OPTIONAL = {
    OUTPUT: dict,
    VARIABLES: dict,
    SWEEP: dict,
}

VALID_SWEEP_MIN = {
    SWEEP_START: [int, float],
    SWEEP_STOP: [int, float],
    SWEEP_JUMP: [int, float]
}

VALID_SWEEP_OPTIONAL = {

}

VALID_RECIPE_MIN = {
    NAME: str,
    RECIPE: dict,
    SOURCE: str
}

VALID_RECIPE_OPTIONAL = {
    ENVIRONMENTS: dict
}

ENVIRONMENTS_MIG = 'mig'
ENVIRONMENTS_LOCAL = 'local'
ENVIRONMENTS_LOCAL_VERSION = 'version'
ENVIRONMENTS_LOCAL_DEPENDENCIES = 'dependencies'

ENVIRONMENTS_MIG_NODES = 'nodes'
ENVIRONMENTS_MIG_CPU_CORES = 'cpu cores'
ENVIRONMENTS_MIG_WALL_TIME = 'wall time'
ENVIRONMENTS_MIG_MEMORY = 'memory'
ENVIRONMENTS_MIG_DISKS = 'disks'
ENVIRONMENTS_MIG_CPU_ARCHITECTURE = 'cpu-architecture'
ENVIRONMENTS_MIG_FILL = 'fill'
ENVIRONMENTS_MIG_ENVIRONMENT_VARIABLES = 'environment variables'
ENVIRONMENTS_MIG_NOTIFICATION = 'notification'
ENVIRONMENTS_MIG_RETRIES = 'retries'
ENVIRONMENTS_MIG_RUNTIME_ENVIRONMENTS = 'runtime environments'

COMPARITORS = [
    '==',
    '>='
]

VALID_ENVIRONMENT_TYPES = [
    ENVIRONMENTS_LOCAL,
    ENVIRONMENTS_MIG
]

VALID_ENVIRONMENTS_MIG = {
    ENVIRONMENTS_MIG_NODES: str,
    ENVIRONMENTS_MIG_CPU_CORES: str,
    ENVIRONMENTS_MIG_WALL_TIME: str,
    ENVIRONMENTS_MIG_MEMORY: str,
    ENVIRONMENTS_MIG_DISKS: str,
    ENVIRONMENTS_MIG_CPU_ARCHITECTURE: str,
    ENVIRONMENTS_MIG_FILL: list,
    ENVIRONMENTS_MIG_ENVIRONMENT_VARIABLES: list,
    ENVIRONMENTS_MIG_NOTIFICATION: list,
    ENVIRONMENTS_MIG_RETRIES: str,
    ENVIRONMENTS_MIG_RUNTIME_ENVIRONMENTS: list
}

VALID_ENVIRONMENTS_MIG_FILLS= [
    'CPUCOUNT',
    'CPUTIME',
    'DISK',
    'MEMORY',
    'NODECOUNT'
]

VALID_ENVIRONMENTS_LOCAL = {
    ENVIRONMENTS_LOCAL_DEPENDENCIES: list
}

VALID_WORKFLOW_MIN = {
    CWL_NAME: str,
    CWL_CWL_VERSION: str,
    CWL_CLASS: str,
    CWL_INPUTS: dict,
    CWL_OUTPUTS: dict,
    CWL_STEPS: dict
}

VALID_WORKFLOW_OPTIONAL = {
    CWL_REQUIREMENTS: dict
}

VALID_STEP_MIN = {
    CWL_NAME: str,
    CWL_CWL_VERSION: str,
    CWL_CLASS: str,
    CWL_BASE_COMMAND: str,
    CWL_INPUTS: dict,
    CWL_OUTPUTS: dict
}

VALID_STEP_OPTIONAL = {
    CWL_STDOUT: str,
    CWL_ARGUMENTS: list,
    CWL_REQUIREMENTS: dict,
    CWL_HINTS: dict
}

VALID_SETTING_MIN = {
    CWL_NAME: str,
    CWL_VARIABLES: dict
}

VALID_SETTING_OPTIONAL = {
}

ANCESTORS = 'ancestors'
DESCENDANTS = 'descendants'
WORKFLOW_INPUTS = 'workflow inputs'
WORKFLOW_OUTPUTS = 'workflow outputs'

CHAR_LOWERCASE = 'abcdefghijklmnopqrstuvwxyz'
CHAR_UPPERCASE = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
CHAR_NUMERIC = '0123456789'
CHAR_LINES = '-_'
CHAR_COMPARISON = '=<>'

NOTEBOOK_EXTENSION = '.ipynb'
NOTEBOOK_EXTENSIONS = [
    NOTEBOOK_EXTENSION
]

GREEN = 'lightgreen'
RED = 'pink'
WHITE = 'lightgray'

COLOURS = [
    GREEN,
    RED,
    WHITE
]

NO_OUTPUT_SET_WARNING = \
    'No output has been set. The recipe may still output data but for the ' \
    'visualisation to be correct it should be told if this is so. '
NO_NAME_SET_ERROR = "A pattern name must be defined. "
NO_INPUT_FILE_SET_ERROR = \
    "An input file must be defined. This is the file that is used to " \
    "trigger any processing and can be defined using the methods " \
    "'.add_single_input' or 'add_gathering_input. "
NO_INPUT_PATH_SET_ERROR = \
    "At least one input path must be defined. This is the path to the file " \
    "that is used to trigger any processing and can be defined using the " \
    "methods '.add_single_input' or 'add_gathering_input. "
INVALID_INPUT_PATH_ERROR = "The input path is not valid. "
NO_RECIPES_SET_ERROR = "No recipes have been defined. "
PLACEHOLDER_ERROR = \
    "A placeholder value was detected. Please update this before proceeding. "

MEOW_MODE = 'MEOW'
CWL_MODE = 'CWL'
WIDGET_MODES = [
    MEOW_MODE,
    CWL_MODE
]

LOGGING_DIR = 'mig_meow_logs'
WORKFLOW_LOGFILE_NAME = 'workfow_widget'
MONITOR_LOGFILE_NAME = 'monitor_widget'
REPORT_LOGFILE_NAME = 'report_widget'
RUNNER_LOGFILE_NAME = 'workflow_runner'

KEYWORD_PATH = "{PATH}"
KEYWORD_REL_PATH = "{REL_PATH}"
KEYWORD_DIR = "{DIR}"
KEYWORD_REL_DIR = "{REL_DIR}"
KEYWORD_FILENAME = "{FILENAME}"
KEYWORD_PREFIX = "{PREFIX}"
KEYWORD_VGRID = "{VGRID}"
KEYWORD_EXTENSION = "{EXTENSION}"
KEYWORD_JOB = "{JOB}"

MIG_TRIGGER_KEYWORDS = [
    KEYWORD_PATH,
    KEYWORD_REL_PATH,
    KEYWORD_DIR,
    KEYWORD_REL_DIR,
    KEYWORD_FILENAME,
    KEYWORD_PREFIX,
    KEYWORD_VGRID,
    KEYWORD_EXTENSION,
    KEYWORD_JOB
]




import ipywidgets as widgets
import ipydatetime

from IPython.display import display
from graphviz import Digraph

from .constants import VGRID_READ, VGRID_REPORT_OBJECT_TYPE
from .logging import create_report_logfile, write_to_log
from .mig import vgrid_report_json_call
from .validation import check_input

DEFAULT_FILENAME = 'workflow_report'
DEFAULT_EXTENSION = 'png'

SUPPORTED_EXTENSIONS = [
    DEFAULT_EXTENSION,
    'pdf'
]


class ReportWidget:
    """

    """
    def __init__(self, vgrid, debug=False, filename=DEFAULT_FILENAME,
                 extension=DEFAULT_EXTENSION, horizontal=False):
        """

        """
        self.logfile = create_report_logfile(debug)

        check_input(vgrid, str, 'vgrid')
        self.vgrid = vgrid

        check_input(filename, str, 'filename')
        self.filename = filename

        check_input(extension, str, 'format')
        if extension not in SUPPORTED_EXTENSIONS:
            raise ValueError(
                "Invalid report extension '%s' given. Supported extensions "
                "are: %s" % (extension, SUPPORTED_EXTENSIONS)
            )
        self.extension = extension

        check_input(horizontal, bool, 'horizontal')
        self.horizontal = horizontal

        self.report = None
        self.jobs = []
        self.paths = []
        self.patterns = []
        self.recipes = []
        self.pattern_filter = None
        self.recipe_filter = None
        self.path_filter = None
        self.job_filter = None
        self.greedy_filter = None
        self.before_filter = None
        self.after_filter = None

        self.header_display_area = widgets.Output()
        self.report_display_area = widgets.Output()

        self.__get_workflow_report()

    def __get_workflow_report(self):
        """

        :return:
        """
        if not self.report:
            valid, result = self.get_vgrid_report()

            if valid:
                self.report = result
                for job_id, job_hist in result.items():
                    self.jobs.append(job_id)
                    pattern = job_hist['pattern_name']
                    if pattern not in self.patterns:
                        self.patterns.append(pattern)
                    recipes = [name for name, _ in job_hist['recipes']]
                    for recipe in recipes:
                        if recipe not in self.recipes:
                            self.recipes.append(recipe)
                    trigger_path = job_hist['trigger_path']
                    if trigger_path not in self.paths:
                        self.paths.append(trigger_path)

                self.jobs.sort()
                self.paths.sort()
                self.patterns.sort()
                self.recipes.sort()

                self.__construct_image()
            else:
                self.__display_error(result)

    def get_vgrid_report(self):
        """

        :return:
        """
        attributes = {}
        _, response, _ = vgrid_report_json_call(
            self.vgrid,
            VGRID_READ,
            VGRID_REPORT_OBJECT_TYPE,
            attributes,
            logfile=self.logfile
        )

        if response['object_type'] == VGRID_REPORT_OBJECT_TYPE:
            report = response['report']
            write_to_log(
                self.logfile,
                'get_vgrid_report',
                'Got report with jobs: %s' % report.keys()
            )
            return True, report
        else:
            msg = 'Something went wrong with retrieving the report. '
            write_to_log(
                self.logfile,
                'get_vgrid_report',
                '%s:%s' % (msg, response)
            )
            if self.logfile:
                msg += "You can check the logfile '%s' for details" \
                       % self.logfile
            return False, msg

    def __display_report(self):
        """

        :return:
        """
        self.__construct_header()
        self.__construct_image()
        self.__load_image()

        widget = widgets.VBox(
            [
                self.header_display_area,
                self.report_display_area
            ],
            layout=widgets.Layout(width='100%')
        )

        return widget

    def __construct_header(self):
        """

        :return:
        """
        contents = []

        self.job_filter = widgets.SelectMultiple(
            options=self.jobs,
            disabled=False,
            layout=widgets.Layout(width='25%')
        )

        self.path_filter = widgets.SelectMultiple(
            options=self.paths,
            disabled=False,
            layout=widgets.Layout(width='25%')
        )

        self.pattern_filter = widgets.SelectMultiple(
            options=self.patterns,
            disabled=False,
            layout=widgets.Layout(width='25%')
        )

        self.recipe_filter = widgets.SelectMultiple(
            options=self.recipes,
            disabled=False,
            layout=widgets.Layout(width='25%')
        )

        contents.append(
            widgets.HBox(
                [
                    widgets.Label(
                        value="Jobs:",
                        layout=widgets.Layout(width='25%')
                    ),
                    widgets.Label(
                        value="Paths:",
                        layout=widgets.Layout(width='25%')
                    ),
                    widgets.Label(
                        value="Patterns:",
                        layout=widgets.Layout(width='25%')
                    ),
                    widgets.Label(
                        value="Recipes:",
                        layout=widgets.Layout(width='25%')
                    )
                ]
            )
        )

        contents.append(
            widgets.HBox(
                [
                    self.job_filter,
                    self.path_filter,
                    self.pattern_filter,
                    self.recipe_filter
                ]
            )
        )

        refresh_button = widgets.Button(
            value=False,
            description='Refresh',
            disabled=False,
            button_style='',
            tooltip='Refresh displayed report',
            layout=widgets.Layout(width='15%', min_width='12ex')
        )
        refresh_button.on_click(self.__refresh_image)

        # greedy_selector = widgets.Checkbox(
        #     value=False,
        #     description='Only direct matches',
        #     disabled=False,
        #     indent=False
        # )

        self.before_filter = ipydatetime.DatetimePicker(
            description='Before:'
        )

        self.after_filter = ipydatetime.DatetimePicker(
            description='After:'
        )

        # first_time = None
        # last_time = None
        # if self.report:
        #     for job_id, job_hist in self.report.items():
        #         copenhagen = pytz.timezone('Europe/Copenhagen')
        #         job_start = parser.parse(job_hist['start'], tzinfo=copenhagen)
        #         if not first_time or first_time < job_start:
        #             first_time = job_start
        #         if not last_time or last_time > job_start:
        #             last_time = job_start
        #
        # if first_time:
        #     self.before_filter.value = first_time
        # if last_time:
        #     self.after_filter.value = last_time

        bottom_row = widgets.HBox([
            self.after_filter,
            self.before_filter,
            # greedy_selector,
            refresh_button
        ])

        contents.append(
             bottom_row
        )

        v_box = widgets.VBox(
            contents
        )

        self.header_display_area.clear_output(wait=True)
        with self.header_display_area:
            display(v_box)

    def __refresh_image(self, *args):
        write_to_log(
            self.logfile,
            '__refresh_image',
            'Refreshing workflow image'
        )

        self.__construct_image()

        write_to_log(
            self.logfile,
            '__refresh_image',
            'Construction complete'
        )

        self.__load_image()

        write_to_log(
            self.logfile,
            '__refresh_image',
            'Completed refresh'
        )

    def __construct_image(self):
        dot = Digraph(comment=self.filename,
                      format=self.extension,
                      node_attr={'shape': 'plaintext'})

        if self.logfile:
            msg = 'filters are:'
            if self.job_filter and self.job_filter.value:
                msg += '\n\tjobs: %s' % str(self.job_filter.value)
            if self.path_filter and self.path_filter.value:
                msg += '\n\tpaths: %s' % str(self.path_filter.value)
            if self.pattern_filter and self.pattern_filter.value:
                msg += '\n\tpatterns: %s' % str(self.pattern_filter.value)
            if self.recipe_filter and self.recipe_filter.value:
                msg += '\n\trecipes: %s' % str(self.recipe_filter.value)
            if self.after_filter and self.after_filter.value:
                msg += '\n\tafter: %s' % str(self.after_filter.value)
            if self.before_filter and self.before_filter.value:
                msg += '\n\tbefore: %s' % str(self.before_filter.value)
            if msg == 'filters are:':
                msg += '\n\t(None)'
            write_to_log(
                self.logfile,
                '__construct_image',
                msg
            )

        if self.report:
            for job_id, job_hist in self.report.items():
                if self.pattern_filter \
                        and self.pattern_filter.value \
                        and job_hist['pattern_name'] \
                        not in self.pattern_filter.value:
                    continue
                if self.recipe_filter \
                        and self.recipe_filter.value\
                        and set(r for r, _ in job_hist['recipes'])\
                        .isdisjoint(self.recipe_filter.value):
                    continue
                if self.path_filter \
                        and self.path_filter.value \
                        and job_hist['trigger_path'] \
                        not in self.path_filter.value:
                    continue
                if self.job_filter \
                        and self.job_filter.value \
                        and job_id not in self.job_filter.value:
                    continue
                if self.before_filter \
                        and self.before_filter.value \
                        and job_hist['start'] > str(self.before_filter.value):
                    continue
                if self.after_filter \
                        and self.after_filter.value \
                        and job_hist['start'] < str(self.after_filter.value):
                    continue

                write_to_log(
                    self.logfile,
                    '__construct_image',
                    'job %s will be displayed' % job_id
                )

                unique_write = []
                for write, _ in job_hist['write']:
                    if write not in unique_write:
                        unique_write.append(write)

                display_string = \
                    "Job: %s" % job_id + \
                    "\nTrigger: %s" % (job_hist['trigger_path']) + \
                    "\nOutput: %s" % unique_write + \
                    "\nPattern: %s" % (job_hist['pattern_name']) + \
                    "\nRecipe: %s" % (
                        [name for name, _ in job_hist['recipes']])
                dot.node(job_id, display_string)

                for child in job_hist['children']:
                    tail_name = dot._quote_edge(job_id)
                    head_name = dot._quote_edge(child)
                    edge = dot._edge % (tail_name, head_name, '')

                    if edge not in dot.body:
                        dot.edge(job_id, child)

                for parent in job_hist['parents']:
                    tail_name = dot._quote_edge(parent)
                    head_name = dot._quote_edge(job_id)
                    edge = dot._edge % (tail_name, head_name, '')

                    if edge not in dot.body:
                        dot.edge(parent, job_id)

        if self.horizontal:
            dot.graph_attr['rankdir'] = 'LR'

        write_to_log(
            self.logfile,
            '__refresh_image',
            'Rendering....'
        )

        dot.render(self.filename)

        write_to_log(
            self.logfile,
            '__refresh_image',
            'Render complete'
        )

    def __load_image(self):
        file = open(self.filename + '.' + self.extension, 'rb')
        image = file.read()
        report = widgets.Image(
                value=image,
                format='.' + self.extension,
        )

        self.report_display_area.clear_output(wait=True)
        with self.report_display_area:
            display(report)

    def __display_error(self, error):
        """

        :param error:
        :return:
        """
        labels = [
            widgets.Label('Could not retrieve workflow job report. %s' % error)
        ]

        label_display = widgets.VBox(labels)

        self.report_display_area.clear_output(wait=True)
        with self.report_display_area:
            display(label_display)

    def display_widget(self):
        """
        Returns the widget in a display ready state.

        :return: (widgets.Output) The output are to be displayed in a notebook.
        """
        return self.__display_report()



import os

from setuptools import setup

current_dir = os.path.abspath(os.path.dirname(__file__))


def read_file(path):
    with open(path, "r") as _file:
        return _file.read()


def read_requirements(filename):
    path = os.path.join(current_dir, filename)
    return [req.strip() for req in read_file(path).splitlines() if req.strip()]


with open('README.md', 'r') as readme:
    long_description = readme.read()

module_data = {}
with open(os.path.join(current_dir, "mig_meow", "version.py")) as f:
    exec(f.read(), {}, module_data)

setup(name=module_data['__name__'],
      version=module_data['__version__'],
      author='David Marchant',
      author_email='d.marchant@ed-alumni.net',
      description='MiG based manager for event oriented workflows',
      long_description=long_description,
      url='https://github.com/PatchOfScotland/mig_meow',
      packages=['mig_meow'],
      install_requires=read_requirements("requirements.txt"),
      extras_require={
                "test": read_requirements("requirements-testing.txt"),
      },
      classifiers=[
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Operating System :: OS Independent'
      ])
